/* YAMPMOD TRACKER: (C) Copyright 1995 David Groves */

/* This program is distributed under the terms of the GNU Public License */
/* Please see the file "COPYING" for details */

/* please contact djg@djghome.demon.co.uk for bug-reports, comments, etc. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/soundcard.h>
#include <fcntl.h>

/* #define NOSOUND 1	*/	/* Debugging only! */

/* ----------------------------------------------------------------------- */
/* Audio low-level stuff...                                                */
/* ----------------------------------------------------------------------- */
int MAXPLAYBUFFER;

int audio_fd;
int frequency;

char *playbuffer;
int  buffer_index;

signed char voltab[64][256];

/* ----------------------------------------------------------------------- */
/* Version and other admin...                                              */
/* ----------------------------------------------------------------------- */
#define VERSION_STRING "0.1"


/* ----------------------------------------------------------------------- */
/* Song information */
/* ----------------------------------------------------------------------- */
struct instrument_tag {
	char name[22];
	int  length;
	int  finetune;
	int  volume;
	int  repstart;
	int  replen;
	/* Information not stored in MOD file... */
	char *sample;
};

struct note_tag {
	/* none of this corresponds to the format of the file, as the actual */
	/* notes are in amiga-packed format. This is unpacked.               */
	int instrument;
	int pitch;
	int effect_cmd;
	int effect_param;
};

struct pattern_tag {
	struct note_tag note[64*4];
};

struct modsong_tag {
	char                  title[20];
	struct instrument_tag instrument[32];
	int                   num_arrangements;
	int                   arrangement[128];
	struct pattern_tag    pattern[64];
	/* Information not stored in MOD file ... */
	int 		      num_patterns;
};

struct modsong_tag song;

/* ----------------------------------------------------------------------- */
/* Period Table                                                            */
/* ----------------------------------------------------------------------- */
char NoteName[48][4] = { 
	"C-1", "C#1", "D-1", "Eb1", "E-1", "F-1", "F#1", "G-1", "Ab1", "A-1", "Bb1", "B-1",
	"C-2", "C#2", "D-2", "Eb2", "E-2", "F-2", "F#2", "G-2", "Ab2", "A-2", "Bb2", "B-2",
	"C-3", "C#3", "D-3", "Eb3", "E-3", "F-3", "F#3", "G-3", "Ab3", "A-3", "Bb3", "B-3",
	"C-4", "C#4", "D-4", "Eb4", "E-4", "F-4", "F#4", "G-4", "Ab4", "A-4", "Bb4", "B-4"
};

int PeriodTab[16][48] = {
/* Tuning 0, Normal */
        { 856,808,762,720,678,640,604,570,538,508,480,453, 
          428,404,381,360,339,320,302,285,269,254,240,226, 
          214,202,190,180,170,160,151,143,135,127,120,113,
	  107,101, 95, 90, 85, 80, 76, 72, 68, 64, 60, 57 },

        { 850,802,757,715,674,637,601,567,535,505,477,450,
          425,401,379,357,337,318,300,284,268,253,239,225,
          213,201,189,179,169,159,150,142,134,126,119,113,
	  106,100, 95, 89, 84, 80, 75, 71, 67, 63, 60, 56 },

        { 844,796,752,709,670,632,597,563,532,502,474,447,
          422,398,376,355,335,316,298,282,266,251,237,224,
          211,199,188,177,167,158,149,141,133,125,118,112,
	  106,100, 94, 89, 84, 79, 75, 70, 67, 63, 59, 56 },

        { 838,791,746,704,665,628,592,559,528,498,470,444,
          419,395,373,352,332,314,296,280,264,249,235,222,
          209,198,187,176,166,157,148,140,132,125,118,111,
	  105, 99, 93, 88, 83, 79, 74, 70, 66, 62, 59, 56 },

        { 832,785,741,699,660,623,588,555,524,495,467,441,
          416,392,370,350,330,312,294,278,262,247,233,220,
          208,196,185,175,165,156,147,139,131,124,117,110,
	  104, 98, 93, 87, 83, 78, 74, 69, 66, 62, 58, 55 },

        { 826,779,736,694,655,619,584,551,520,491,463,437,
          413,390,368,347,328,309,292,276,260,245,232,219,
          206,195,184,174,164,155,146,138,130,123,116,109,
	  103, 97, 92, 87, 82, 77, 73, 69, 65, 61, 58, 55 },

        { 820,774,730,689,651,614,580,547,516,487,460,434,
          410,387,365,345,325,307,290,274,258,244,230,217,
          205,193,183,172,163,154,145,137,129,122,115,109,
	  103, 97, 91, 86, 81, 77, 73, 68, 65, 61, 58, 54 },

        { 814,768,725,684,646,610,575,543,513,484,457,431,
          407,384,363,342,323,305,288,272,256,242,228,216,
          204,192,181,171,161,152,144,136,128,121,114,108,
	  102, 96, 91, 86, 81, 76, 72, 68, 64, 61, 57, 54 },

        { 907,856,808,762,720,678,640,604,570,538,508,480,
          453,428,404,381,360,339,320,302,285,269,254,240,
          226,214,202,190,180,170,160,151,143,135,127,120,
	  113,107,801, 95, 90, 85, 80, 76, 71, 67, 64, 60 },

        { 900,850,802,757,715,675,636,601,567,535,505,477,
          450,425,401,379,357,337,318,300,284,268,253,238,
          225,212,200,189,179,169,159,150,142,134,126,119,
	  113,106,100, 95, 89, 84, 80, 75, 71, 67, 63, 60 },

        { 894,844,796,752,709,670,632,597,563,532,502,474,
          447,422,398,376,355,335,316,298,282,266,251,237,
          223,211,199,188,177,167,158,149,141,133,125,118,
	  112,106,100, 94, 89, 84, 79, 75, 70, 67, 63, 60 },

        { 887,838,791,746,704,665,628,592,559,528,498,470,
          444,419,395,373,352,332,314,296,280,264,249,235,
          222,209,198,187,176,166,157,148,140,132,125,118,
	  111,105, 99, 93, 88, 83, 79, 74, 70, 66, 62, 59 },

        { 881,832,785,741,699,660,623,588,555,524,494,467,
          441,416,392,370,350,330,312,294,278,262,247,233,
          220,208,196,185,175,165,156,147,139,131,123,117,
	  110,104, 98, 93, 87, 83, 78, 74, 69, 66, 62, 58 },

        { 875,826,779,736,694,655,619,584,551,520,491,463,
          437,413,390,368,347,328,309,292,276,260,245,232,
          219,206,195,184,174,164,155,146,138,130,123,116,
	  109,103, 97, 92, 87, 82, 77, 73, 69, 65, 61, 58 },

        { 868,820,774,730,689,651,614,580,547,516,487,460,
          434,410,387,365,345,325,307,290,274,258,244,230,
          217,205,193,183,172,163,154,145,137,129,122,115,
	  109,103, 97, 91, 86, 81, 77, 73, 68, 65, 61, 58 },

        { 862,814,768,725,684,646,610,575,543,513,484,457,
          431,407,384,363,342,323,305,288,272,256,242,228,
          216,203,192,181,171,161,152,144,136,128,121,114,
	  108,102, 96, 91, 86, 81, 76, 72, 68, 64, 61, 57 }
};

/* ----------------------------------------------------------------------- */
/* Vibrato Table                                                           */
/* ----------------------------------------------------------------------- */

/* sine table for the vibrato effect (obtained through build.c) */

int vibratotab[64] = 
   {
   0,50,100,149,196,241,284,325,362,396,426,452,473,490,502,510,512,
   510,502,490,473,452,426,396,362,325,284,241,196,149,100,50,0,-49,
   -99,-148,-195,-240,-283,-324,-361,-395,-425,-451,-472,-489,-501,
   -509,-511,-509,-501,-489,-472,-451,-425,-395,-361,-324,-283,-240,
   -195,-148,-99,-49
   };

/* ----------------------------------------------------------------------- */
/* Player state information                                                */
/* ----------------------------------------------------------------------- */

int pattern_number;
int arrangement_number;
int note_number;
int samples_per_tick;
int speed;
int finespeed;
int countdown;		/* countdown until we need to advance another note */
int tickdown;
int show_playing = 0;
int skip_set = 0;
int skip_note;
int skip_arrangement;
int delay_pattern = 0;
int delay_note    = 0;

struct channel_tag {
	int is_deactivated;	/* 0 = plays as per normal! */
	int instrument;
	int position;
	int oldposition;
	int volume;
	int step;
	/* no doubt we will need more for the various effects... */
	int status;
	int note;
	int finetune;
	int pitch;
	int oldpitch;
	char *sample;
	int length;
	int repstart; /* in bytes! */
	int replen;
	int fixlength;
	int fixreplen;
	int fixrepstart;
	int effect;
	int oldeffect;
	int param1;	/* these two number are Lo/Hi of param word */
	int param2;
	int param;	/* param is the effect word as in mod file */
	int volumerate;
	int pitchrate;
	int destpitch;
	int vibratorate;
	int vibratodepth;
	int vibratooffset;
};

struct channel_tag channel[4];

/* ----------------------------------------------------------------------- */
/* The Program Itself...                                                   */
/* ----------------------------------------------------------------------- */

void init_voltab( void )
{
	int i;
	int j;

	for( i = 0; i < 256; i++ ) {
		for( j = 0; j < 64; j++ ) {
			voltab[j][i] = (signed char) (((i-128)*j)/256);
		}
	}

}
void open_audio( void )
{
	int samplesize;
	int stereo;

	audio_fd = open("/dev/dsp", O_WRONLY | O_SYNC, 0 );
	if( audio_fd == -1 ) {
		fprintf( stderr, "Unable to open sound device!\n" );
		exit(1);
	}

	samplesize = 8;
	if( ioctl( audio_fd, SNDCTL_DSP_SAMPLESIZE, &samplesize ) == -1 ) {
		fprintf( stderr, "Unable to set sample size!\n");
		exit(1);
	}

	stereo = 0;
	if( ioctl( audio_fd, SNDCTL_DSP_STEREO, &stereo ) == -1 ) {
		fprintf( stderr, "Unable to set mono playback!\n");
		exit(1);
	}

	frequency = 22050;
	if( ioctl( audio_fd, SNDCTL_DSP_SPEED, &frequency ) == -1 ) {
		fprintf( stderr, "Unable to set playback frequency!\n");
		exit(1);
	}
	if( ioctl( audio_fd, SNDCTL_DSP_GETBLKSIZE, &MAXPLAYBUFFER ) == -1 ) {
		fprintf( stderr, "Unable to get block size!\n");
		exit(1);
	}
	MAXPLAYBUFFER = MAXPLAYBUFFER * 4;
	printf("\nBlock size = %d\n", MAXPLAYBUFFER );

	playbuffer = malloc( MAXPLAYBUFFER );
	memset( playbuffer, 128, MAXPLAYBUFFER );
	
	buffer_index = 0;
}

void write_buffer( void )
{
	write( audio_fd, playbuffer, buffer_index );
	buffer_index = 0;
/* 	printf("\nWRITING BUFFER...\n"); */
}

void close_audio( void )
{	
	write_buffer();
	close( audio_fd );
	free( playbuffer );
}

void write_sample( unsigned char sample )
{
	playbuffer[buffer_index++] = sample;
	if( buffer_index >= MAXPLAYBUFFER ) {
		write_buffer();
	}
}

char *load_bytes( FILE *f, int len )
{
	char *retval;

	retval = malloc( len );
	if( retval == NULL ) {
		fprintf(stderr, "Out of memory!\n" );
		exit(1);
	}
	fread( retval, 1, len, f );

	if( feof( f ) ) {
		fprintf(stderr, "Unexpected end-of-file!\n");
		exit(1);
	}
	return retval;
}

unsigned char load_byte( FILE *f )
{
	unsigned char c;

	fread( &c, 1, 1, f );
	if( feof( f ) ) {
		fprintf( stderr, "Unexpected end-of-file!\n");
		exit(1);
	}

	return c;
}

int load_short( FILE *f )
{
	int retval;

	retval = load_byte( f ) << 8;
	retval |= load_byte( f );
	return retval;
}

void getsong_title( FILE *f )
{
	char *title;
	title = load_bytes(f, 20 ); 	
	strcpy( song.title, title );
	printf("Song: %s\n", title );
	free( title );
}

void getsong_instruments( FILE *f )
{
	int i;
	char *insname;
	
	for( i = 1; i < 32; i++ ) {
		insname = load_bytes( f, 22 );
		strcpy( song.instrument[i].name, insname );
		song.instrument[i].length = load_short( f );
		song.instrument[i].finetune = load_byte( f );
		if( song.instrument[i].finetune > 15 ) {
			song.instrument[i].finetune = 0; /* why ??? */
		}
		song.instrument[i].volume = load_byte( f );
		if( song.instrument[i].volume > 63 ) {
			song.instrument[i].volume = 63;
		}
		song.instrument[i].repstart = load_short( f );
		song.instrument[i].replen   = load_short( f );
		if( song.instrument[i].repstart == 0 )
			song.instrument[i].replen = 0;

		/* Initialise sample (haven't read it yet... */
		song.instrument[i].sample   = NULL;

		if( song.instrument[i].length > 0 ) {
			printf(" I%02X: %22s %5d %2d %2d %5d %5d \n", i, insname,
				song.instrument[i].length,
				song.instrument[i].finetune,
				song.instrument[i].volume,
				song.instrument[i].repstart,
				song.instrument[i].replen );
		}

		free( insname );
	}
}

int getsong_numberofpatterns( FILE *f )
{
	int i;
	int maxpat;

	/* at present, we cheat, by calculating the highest pattern number */
	/* in the arrangement. Other MOD players do this also, but I don't */
	/* think it's always correct... */

	maxpat = -1;
	for( i = 0; i < song.num_arrangements; i++ ) {
		if( song.arrangement[i] > maxpat ) {
			maxpat = song.arrangement[i];
		}
	}
	maxpat++;
	if( maxpat == 0 ) {
		fprintf( stderr, "Unable to determine maximum pattern!\n" );
		exit(1);
	}

	return maxpat;
}
void getsong_loadnote( FILE *f, int pnumber, int nnumber )
{
	int a;
	int b;
	int c;
	int d;

	a = load_byte( f );
	b = load_byte( f );
	c = load_byte( f );
	d = load_byte( f );

	song.pattern[pnumber].note[nnumber].instrument   = (a & 0x10)|(c >> 4);
	song.pattern[pnumber].note[nnumber].effect_cmd   = c & 0x0f;
	song.pattern[pnumber].note[nnumber].effect_param = d;
	song.pattern[pnumber].note[nnumber].pitch        = ((a & 0x0f) << 8)|b;
}

void getsong_loadpattern(FILE *f, int pnumber )
{
	int i;
	/* load a pattern from current position in file into pattern number */
	/* pnumber.                                                         */

	for( i = 0; i < 64; i++ ) {
		getsong_loadnote(f, pnumber, 0+(i*4) );
		getsong_loadnote(f, pnumber, 1+(i*4) );
		getsong_loadnote(f, pnumber, 2+(i*4) );
		getsong_loadnote(f, pnumber, 3+(i*4) );
	}
}

void getsong_patterns( FILE *f )
{
	int dummy;
	int i;
	char file_marker[5];
	char *file_marker1;

/* 	printf( "DEBUG: currently at byte %x\n", ftell( f ) ); */

	song.num_arrangements = load_byte( f );

	if( song.num_arrangements == 0 ) {
		fprintf(stderr, "No arrangements found!\n" );
		exit(1);
	}

	/* This byte is meaningless on a PC */
	dummy = load_byte( f );

	printf(" Number of patterns to play: %d\n", song.num_arrangements );


	for( i = 0; i < 128; i++ ) {
		song.arrangement[i] = load_byte( f );
		if( i < song.num_arrangements ) { 
			/* printf(" P%03d: %2d\n", i, song.arrangement[i] ); */
		}
	}

	/* We have now loaded the arrangement successfully. Now come the */
	/* patterns... 							 */

	/* But first, we should find a file type in here... */
	file_marker1 = load_bytes( f, 4 );
	file_marker[0] = file_marker1[0];
	file_marker[1] = file_marker1[1];
	file_marker[2] = file_marker1[2];
	file_marker[3] = file_marker1[3];
	file_marker[4] = 0;

	printf("File marker: %s\n", file_marker );

	free( file_marker1 );

	song.num_patterns = getsong_numberofpatterns( f );
	printf("Number of patterns: %d\n", song.num_patterns );

	/* Now we know how many there are, we can load 'em! */

	for( i = 0; i < song.num_patterns; i++ ) {
		getsong_loadpattern(f, i );
	}
	
}

void getsong_samples( FILE *f )
{
	int i;
	int slength;

	for( i = 1; i < 32; i++ ) {
		slength = song.instrument[i].length;
		if( slength == 0 )
			continue;
		song.instrument[i].sample = load_bytes( f, slength*2);
	}
}

void load_song( char *filename )
{
	FILE *songfile;

	songfile = fopen( filename, "r" );
	if( songfile == NULL ) {
		fprintf( stderr, "Unable to load song %s\n", filename );
		exit(1);
	}
	getsong_title( songfile );
	getsong_instruments( songfile );
	getsong_patterns( songfile );	
	getsong_samples( songfile );
}

int calc_step( int note, int finetune )
{
	if( frequency == 22050 )
		return ( 10541918 / PeriodTab[finetune][note]);
	else 
		return (  5270959 / PeriodTab[finetune][note]);
}

int calc_note( int pitch )
{
	int loop;
	int gotit;

	loop = 0;
	gotit = 0;
	while (loop < 48 ) {
		if( pitch >= PeriodTab[0][loop] ) {
			gotit = 1;
			break;
		}
		loop++;
	}
	if( gotit == 0 ) {
		printf("No note matched [%d]!\n", pitch);
		return -1;
	}
	return loop;
}

int set_effect( int c )
{
	switch( channel[c].effect ) {
		case 0:
			if( channel[c].param != 0 ) {
				printf("Arpeggio not supported!\n");
			}
			break;
		case 1:
			channel[c].pitchrate = -channel[c].param;
			break;
		case 2:
			channel[c].pitchrate = channel[c].param;
			break;
		case 3:
			/* portamento to note: slide from oldpitch to pitch */
			if( channel[c].pitch > channel[c].oldpitch ) {
				/* goin' up... */
				channel[c].pitchrate = channel[c].param;
			}
			else {
				/* goin' down... */
				channel[c].pitchrate = -channel[c].param;
			}
			channel[c].destpitch = channel[c].pitch;
			channel[c].pitch     = channel[c].oldpitch;
			channel[c].step   = 10541918 / channel[c].pitch;
			channel[c].position = channel[c].oldposition;
			break;
			
		case 4:
			channel[c].vibratodepth = channel[c].param1;
			channel[c].vibratorate  = channel[c].param2;
			if( (channel[c].oldeffect != 4) && (channel[c].oldeffect != 6 ) ) {
				channel[c].vibratooffset = 0;
			}
			break;
		case 5:
			/* continue tone-portamento (effect 3), and slide volume */
			if( channel[c].param1 > 0 ) {
				channel[c].volumerate = -channel[c].param1;
			}
			else {
				channel[c].volumerate = channel[c].param2;
			}
			break;

		case 6:
			/* volume slide, continuing vibrato */
			if( channel[c].param1 > 0 ) {
				channel[c].volumerate = -channel[c].param1;
			}
			else {
				channel[c].volumerate = channel[c].param2;
			}
			break;
		case 9:
			/* sample offset */
			channel[c].position = channel[c].param * 256 * 65536;
			break;
		case 0x0a:
			if( channel[c].param1 > 0 ) {
				channel[c].volumerate = -channel[c].param1;
			}
			else {
				channel[c].volumerate = channel[c].param2;
			}
			break;
		case 0x0c:
			channel[c].volume = channel[c].param;
			break;
		case 0x0b:
			skip_set = 1;
			skip_note = 0;
			skip_arrangement = channel[c].param;
			break;
		case 0x0d:
			skip_set = 1;
			skip_note = channel[c].param1 + (10*channel[c].param2);
			skip_arrangement = arrangement_number +1;
			break;
		case 0x0e:
			switch( channel[c].param2 ) {
				case 0x00:
					printf("Filter not supported on PC!\n");
					break;
				case 0x03:
					printf("Pattern looping not supported on PC!\n");
					break;
				case 0x09:
					printf("Regtrigger note is not supported!\n");
					break;
				case 0x0a:
					channel[c].volume+=channel[c].param1;
					if( channel[c].volume >= 63 )
						channel[c].volume = 63;
					break;
				case 0x0b:
					channel[c].volume-=channel[c].param1;
					if( channel[c].volume <= 0 )
						channel[c].volume = 0;
					break;
				case 0x0d:
					printf("Experimental: delay note %X\n", channel[c].param1 );
					delay_note = channel[c].param1;
					break;
				case 0x0e:
					printf("Experimental: delay pattern %X\n", channel[c].param1 );
					delay_pattern = channel[c].param1;
					break;
				default:
					printf("Effect E [%X (%X)] not supported!\n", 
						channel[c].param2, channel[c].param1 );
					break;
			}
			break;		
		case 0x0f:
			if( channel[c].param > 31 ) {
				finespeed = channel[c].param;
				samples_per_tick = (finespeed*(frequency*15)/750)/125;
				printf("Finespeed set: %d\n", channel[c].param );
			}
			else
				speed = channel[c].param;
			break;
		default:
			printf("Command %X not supported!\n", channel[c].effect );
			break;
	}
	return 0;
}

void set_note( void )
{
	int i;
	static struct note_tag n[4];

restart:;
	
	memcpy( &n[0], 
		&song.pattern[pattern_number].note[note_number*4], 
		sizeof( struct note_tag ) * 4 );

	if( show_playing == 1 ) {
		/* Display Note */
		for( i = 0; i < 4; i++ ) {
			int note;
			if( n[i].pitch != 0 ) {
				note = calc_note( n[i].pitch );
			}
			else {
				note = -1;
			}
			if( note == -1 ) {
				printf("| %02X --- %02X %02X ", n[i].instrument, n[i].effect_cmd,
							   n[i].effect_param );
			}
			else {
				printf("| %02X %s %02X %02X ", n[i].instrument, NoteName[note], n[i].effect_cmd,
							n[i].effect_param );
			}
		}
		printf("|\n");
	}

	for( i = 0; i < 4; i++ ) {
		if( channel[i].is_deactivated == 1 )
			continue;
		channel[i].oldeffect = channel[i].effect;
		channel[i].effect     = n[i].effect_cmd;
		channel[i].param1     = n[i].effect_param & 0x0f;
		channel[i].param2     = n[i].effect_param >> 4;
		channel[i].param      = n[i].effect_param;
		channel[i].oldposition   = channel[i].position;
		channel[i].oldpitch   = channel[i].pitch;
		if( n[i].instrument != 0 ) { 
			if( song.instrument[n[i].instrument].sample == NULL ) {
				/* channel[i].status = 0; */
				goto not_this_instrument;
			}
			channel[i].status = 1;
			channel[i].instrument = n[i].instrument;
			channel[i].volume     = song.instrument[n[i].instrument].volume;
			channel[i].sample     = song.instrument[n[i].instrument].sample;
			channel[i].length     = song.instrument[n[i].instrument].length*2;
			channel[i].repstart   = song.instrument[n[i].instrument].repstart*2;
			channel[i].replen     = song.instrument[n[i].instrument].replen*2; 
			channel[i].finetune   = song.instrument[n[i].instrument].finetune;
			channel[i].fixreplen  = channel[i].replen << 16;
			channel[i].fixlength  = channel[i].length << 16;
			channel[i].fixrepstart = channel[i].repstart << 16;
		}
not_this_instrument:;
		if( ( n[i].pitch != 0 ) &&( channel[i].sample != NULL ) ) {
			channel[i].note       = calc_note( n[i].pitch );
			if( channel[i].note == -1 ) {
				goto not_this_note;
			}
			channel[i].step       = calc_step( channel[i].note, channel[i].finetune );
			if( channel[i].is_deactivated == 2 ) {
				printf("STEP = %d", channel[i].step );
			}
			channel[i].pitch      = PeriodTab[channel[i].finetune][channel[i].note];
			channel[i].position   = 0;
		}
not_this_note:;

		if( set_effect( i )== 1 )
			goto restart;
		tickdown = speed;
	}
}

void do_effects( void )
{
	int i;
	int offset;

	for( i = 0; i < 4; i++ ) {
		switch( channel[i].effect ) {
		    case 1:
		    case 2:
			channel[i].pitch += channel[i].pitchrate;
			channel[i].step   = 10541918 / channel[i].pitch;
			break;
		    case 3:
			channel[i].pitch += channel[i].pitchrate;
			if( channel[i].pitchrate > 0  )	{
				if( channel[i].pitch >= channel[i].destpitch ) {
					channel[i].pitch = channel[i].destpitch;
					channel[i].effect = 0;
				}
			}
			else {
				if( channel[i].pitch <= channel[i].destpitch ) {
					channel[i].pitch = channel[i].destpitch;
					channel[i].effect = 0;
				}

			}
			channel[i].step   = 10541918 / channel[i].pitch;
			break;
		    case 4:
			channel[i].vibratooffset += channel[i].vibratorate;
			channel[i].vibratooffset &= 63;
			offset = (vibratotab[channel[i].vibratooffset] *
					channel[i].vibratodepth ) / 256;
			channel[i].step   = 10541918 / (channel[i].pitch + offset);
			break;
		    case 5:
			channel[i].volume += channel[i].volumerate;
			if( channel[i].volume > 63 )
				channel[i].volume = 63;
			if( channel[i].volume <= 0 )
				channel[i].volume = 0;
			channel[i].pitch += channel[i].pitchrate;
			if( channel[i].pitchrate > 0  )	{
				if( channel[i].pitch >= channel[i].destpitch ) {
					channel[i].pitch = channel[i].destpitch;
					channel[i].effect = 0x0a; /* having hit the */
						/* correct pitch, it's just a */
						/* normal volume-slide... */
				}
			}
			else {
				if( channel[i].pitch <= channel[i].destpitch ) {
					channel[i].pitch = channel[i].destpitch;
					channel[i].effect = 0x0a;
				}

			}
			channel[i].step   = 10541918 / channel[i].pitch;

		    case 6:
			channel[i].volume += channel[i].volumerate;
			if( channel[i].volume > 63 )
				channel[i].volume = 63;
			if( channel[i].volume <= 0 )
				channel[i].volume = 0;
			channel[i].vibratooffset += channel[i].vibratorate;
			channel[i].vibratooffset &= 63;
			offset = (vibratotab[channel[i].vibratooffset] *
					channel[i].vibratodepth ) / 256;
			channel[i].step   = 10541918 / (channel[i].pitch + offset);
			break;
		    case 0x0a:
			channel[i].volume += channel[i].volumerate;
			if( channel[i].volume > 63 )
				channel[i].volume = 63;
			if( channel[i].volume <= 0 )
				channel[i].volume = 0;
			break;
		}
	}
}

signed char buffer_values[8192];

void resample( int samples_to_do )
{
	int value;
	int i;
	int soffset;
	char *sample;
	int j;
	int volume;
	int step;
	int replen;
	int fixreplen;
	int fixreptrigger;
	int fixlength;
	signed char *tab;
	/* the biz! */

	memset( &buffer_values[0], 128, samples_to_do );

	for( i = 0; i < 4; i++ ) {

		if( channel[i].status == 0 )
			continue;

		soffset = channel[i].position;
		sample = channel[i].sample;
		volume = channel[i].volume;
		step   = channel[i].step;
		replen = channel[i].replen;
		fixreplen   = channel[i].fixreplen;
		fixreptrigger = channel[i].fixrepstart + fixreplen;
		fixlength    = channel[i].fixlength;
		tab = (&voltab[volume][0])+128;

		if( replen >= 2 ) {

	            for( j = 0; j < samples_to_do; j++ ) {

			/* value = voltab[volume][(sample[soffset >> 16])+128];  */
			value = *(tab + (sample[soffset >> 16])); 
			soffset += step;

			while( soffset >= fixreptrigger ) { 
				soffset -= fixreplen; 
			}
			buffer_values[j] += value;
		    }	
		}
		else {

	            for( j = 0; j < samples_to_do; j++ ) {
			/* value = voltab[volume][(sample[soffset >> 16])+128];  */
			value = *(tab + (sample[soffset >> 16])); 
			soffset += step;

			if( soffset >= fixlength )
				channel[i].status = 0;
			buffer_values[j] += value;
		    }
		}
		channel[i].position = soffset;
	}
#ifndef NOSOUND
	write( audio_fd, buffer_values, samples_to_do );
#endif
/*	for( j = 0; j < samples_to_do; j++ ) */
		/* write_sample( buffer_values[j] ); */

}

int get_pattern( int arrangement )
{
	arrangement_number = arrangement;
	printf("--- PATTERN: Arr %d ---\n", arrangement);
	if( arrangement_number >= song.num_arrangements )
		return 0;

	pattern_number = song.arrangement[arrangement_number];
	return 1;
}

int get_next_pattern( void )
{
	arrangement_number++;
	printf("--- PATTERN: Arr %d ---\n", arrangement_number);
	if( arrangement_number >= song.num_arrangements )
		return 0;

	pattern_number = song.arrangement[arrangement_number];
	note_number = 0;
	return 1;
}

int get_next_note( void )
{
	int retval;
	if( delay_note != 0 ) {
		set_note();
		delay_note--;
		return 1;
	}
	if( skip_set == 1 ) {
		if( get_pattern( skip_arrangement ) == 0 )
			return 0;
		note_number = skip_note;
		skip_set = 0;
		skip_arrangement = 0;
		skip_note = 0;
		if( delay_pattern == 0 ) {
			set_note();
		}
		return 1;
	}
	if( note_number >= 63 ) {
		retval = get_next_pattern();
		if( delay_pattern == 0 ) {
			set_note();
		}
		return retval;
	}
	else {
		if( delay_pattern != 0 ) {
			delay_pattern--;
		}
		else {
			note_number++;
			set_note();
		}
	}
	return 1;
}

void play_mod( void )
{
	memset( &channel[0], 0, sizeof( struct channel_tag ) * 4 );

	channel[0].is_deactivated = 0;
	channel[1].is_deactivated = 0;
	channel[2].is_deactivated = 0;
	channel[3].is_deactivated = 0;
	finespeed = 125;
	samples_per_tick = (finespeed*(frequency*15)/750)/125;
	speed = 6;
	tickdown = speed;
	arrangement_number = 0;
	pattern_number = song.arrangement[arrangement_number];
	note_number = 0;
	skip_set = 0;
	set_note();

next_tick:;
	countdown = samples_per_tick;
	resample( countdown );
	countdown = 0;
	do_effects();
	tickdown--;
	if( tickdown == 0 ) {
		if( get_next_note() == 0 ) {
			return;
		}
		tickdown = speed;
	}
	goto next_tick;
}

void play_sample( int number )
{
	int length;
	int i;
	char *samp;

	length = song.instrument[number].length;
	samp   = song.instrument[number].sample;
	if( samp == NULL )
		return;
	for( i = 0; i < length; i++ ) {
		write_sample( (*samp++)+128 );
	}
}

void play_samples( void )
{
	int i;

	for( i = 1; i < 32; i++ ) {
		play_sample(i);
	}
}

void main( int argc, char *argv[] )
{
	printf("\nyampmod version %s\n", VERSION_STRING );	
	if( argc == 1 ) {
		printf("\nmytracker [-v] <filename>\n");
		exit(1);
	}
	else if (argc == 2 ) {
		optind = 1;
	}
	else if (argc == 3 ) {
		optind = 2;
		show_playing = 1;
	}
	load_song( argv[optind] );
	init_voltab();
	open_audio();	
/*	play_samples();		*/
	play_mod();
	close_audio();
}


